﻿using System;
using System.Collections.Generic;
using System.IO;
using System.Diagnostics;
using System.Windows.Forms;
using System.Drawing;
using GustFront;
using SharpDX.Direct3D9;
using SharpDX.XACT3;
using SharpDX.XInput;
using SharpDX.Windows;

namespace RVII
{
    public sealed class RVIIManager : ScriptManager
    {
        private static RVIIManager globalInstance = null;

        private RenderForm myForm = null;
        private Direct3D myD3D = null;
        private Device myDev = null;
        private Sprite mySprite = null;
        private Texture myPlaneTex = null;
        private Texture myShaderTex = null;
        private Texture myWndTex = null;
        private AudioEngine myAudioEngine = null;
        private SoundBank mySoundBank = null;
        private System.Drawing.Text.PrivateFontCollection myPFC = null;
        private PresentParameters myD3DPP;
        private long myPrevTick = 0L;
        private bool myDevMustReset = false;

        public RVIIManager(UnregisterActionCallback unreg)
            : base(unreg)
        {
            globalInstance = this;
        }

        public static RVIIManager Global
        {
            get { return globalInstance; }
        }

        internal Form Window { get { return myForm; } }
        internal Device GraphicsDevice { get { return myDev; } }
        internal Texture ShaderTex { get { return myShaderTex; } }
        internal Texture WndTex { get { return myWndTex; } }
        internal Texture PlaneTex { get { return myPlaneTex; } }
        internal AudioEngine AudioEngine { get { return myAudioEngine; } }
        internal SoundBank SoundBank { get { return mySoundBank; } }

        public Rectangle GetViewBounds()
        {
            SharpDX.Viewport vp = myDev.Viewport;
            return new Rectangle(vp.X, vp.Y, vp.Width, vp.Height);
        }

        internal bool Save(int slotNumber, string typeName, string fileName)
        {
            string filePath = Util.CombinePath(Environment.CurrentDirectory, "Data",
                String.Format("Slot{0:00}", slotNumber), fileName);
            using (StreamWriter writer = new StreamWriter(GetStreamToWrite(filePath), TextEncoding)) {
                Element.Save(writer, typeName);
            }
            return true;
        }

        internal Element[] Load(int slotNumber, string typeName, string fileName)
        {
            string filePath = Util.CombinePath(Environment.CurrentDirectory, "Data",
                String.Format("Slot{0:00}", slotNumber), fileName);
            using (StreamReader reader = new StreamReader(GetStreamToRead(filePath), TextEncoding)) {
                IHasElement he = globalInstance.GetPluginElementInterface(typeName);
                if (he != null) {
                    return he.LoadElement(reader, typeName);
                } else if (Util.CaseInsensitiveEquals(typeName, "Texture")) {
                    return Element.Load(reader, typeName, typeof(TextureInfo));
                } else if (Util.CaseInsensitiveEquals(typeName, "Shader")) {
                    Element[] elms = Element.Load(reader, typeName, typeof(ShaderInfo));
                    foreach (ShaderInfo t in elms) {
                        t.Instance = Effect.FromStream(myDev, GetStreamToRead(t.GetData("Path").AsString()), ShaderFlags.None);
                    }
                    return elms;
                } else if (Util.CaseInsensitiveEquals(typeName, "WaveBank")) {
                    return Element.Load(reader, typeName, typeof(WaveInfo));
                } else {
                    return Element.Load(reader, typeName, typeof(UserElement));
                }
            }
        }

        private sealed class BootEngine : RVIIEngine
        {
            public BootEngine(UnregisterActionCallback unreg, ScriptEngine parent, ModuleInfo script, params ScriptValue[] args)
                : base(unreg, script, args)
            {
            }

            #region "cmd impl"

            protected override IDictionary<string, CommandImpl> GetCommandMap()
            {
                Dictionary<string, CommandImpl> dic = new Dictionary<string, CommandImpl>(base.GetCommandMap());

                return dic;
            }

            #endregion

            public override void Render(IGraphicsDevice r)
            {
                RenderAllProcess(r);
                base.Render(r);
            }

            public override void Tick(TimeSpan delta)
            {
                TickAllProcess(delta);
                base.Tick(delta);
            }
        }

        private void EndOfTheGame(Action obj)
        {
            foreach (Action anAction in GetRegisteredActions()) {
                if (anAction != obj) {
                    anAction.EndOfAction();
                }
            }
            UnregisterAction(obj);
        }

        public override void Fire()
        {
            base.Fire();

            Element rvii = Element.Load(new StreamReader(GetStreamToRead("RVII.ini"), TextEncoding), "RVII", typeof(UserElement))[0];

            myForm = new RenderForm(rvii.GetData("Title").AsString());
            myForm.MaximizeBox = false;
            myForm.FormBorderStyle = FormBorderStyle.FixedSingle;
            myForm.ClientSize = new System.Drawing.Size(rvii.GetData("ScreenWidth").AsInt32(), rvii.GetData("ScreenHeight").AsInt32());

            myForm.Show();

            myD3D = new Direct3D();

            myD3DPP.BackBufferWidth = rvii.GetData("ScreenWidth").AsInt32();
            myD3DPP.BackBufferHeight = rvii.GetData("ScreenHeight").AsInt32();
            myD3DPP.BackBufferFormat = Format.Unknown;
            myD3DPP.BackBufferCount = 1;
            myD3DPP.AutoDepthStencilFormat = Format.Unknown;
            myD3DPP.FullScreenRefreshRateInHz = 0;
            myD3DPP.MultiSampleType = MultisampleType.None;
            myD3DPP.MultiSampleQuality = 0;
            myD3DPP.PresentationInterval = PresentInterval.Default;
            myD3DPP.PresentFlags = PresentFlags.None;
            myD3DPP.SwapEffect = SwapEffect.Copy;
            myD3DPP.Windowed = true;
            myD3DPP.DeviceWindowHandle = myForm.Handle;

            myDev = new Device(myD3D, 0, SharpDX.Direct3D9.DeviceType.Hardware, myForm.Handle,
                CreateFlags.HardwareVertexProcessing, myD3DPP);

            PrepareD3DResource();

            AudioEngineSettings aes = new AudioEngineSettings(GetStreamToRead(rvii.GetData("AudioEngine").AsString()));
            myAudioEngine = new AudioEngine(CreationFlags.None, aes);
            mySoundBank = new SoundBank(myAudioEngine, GetStreamToRead(rvii.GetData("SoundBank").AsString()));

            InputUtil.RegisterKeyboard();

            InputUtil.MapButton(44, 0);
            InputUtil.MapButton(43, 1);
            InputUtil.MapButton(42, 2);
            InputUtil.MapButton(41, 3);
            InputUtil.MapButton(1, 4);
            InputUtil.MapButton(2, 5);
            InputUtil.MapButton(3, 6);
            InputUtil.MapButton(4, 7);

            InputUtil.MapKeyToButton((int)Keys.Left, 41);
            InputUtil.MapKeyToButton((int)Keys.Right, 42);
            InputUtil.MapKeyToButton((int)Keys.Up, 43);
            InputUtil.MapKeyToButton((int)Keys.Down, 44);
            InputUtil.MapKeyToButton((int)Keys.Z, 1);
            InputUtil.MapKeyToButton((int)Keys.X, 2);
            InputUtil.MapKeyToButton((int)Keys.A, 3);
            InputUtil.MapKeyToButton((int)Keys.S, 4);

            myPFC = new System.Drawing.Text.PrivateFontCollection();
            foreach (string font_name in ((System.Collections.IList)rvii.GetData("Fonts").Instance)) {
                UserElement font_data = (UserElement)Element.GetElement(font_name);
                string font_file_name = font_data.GetData("Path").AsString();
                if (File.Exists(font_file_name)) {
                    myPFC.AddFontFile(font_file_name);
                } else {
                    myPFC.AddFontFile(Util.CombinePath(Environment.GetEnvironmentVariable("SYSTEMROOT"), "Fonts", font_file_name));
                }
                string name = font_data.GetData("Name").AsString();
                float size = font_data.GetData("Size").AsSingle();
                int style = font_data.GetData("Style").AsInt32();
                foreach (FontFamily ff in myPFC.Families) {
                    if (Util.CaseInsensitiveEquals(ff.Name, name)) {
                        FontInfo fi = new FontInfo(font_data.Tag);
                        fi.Instance = new System.Drawing.Font(ff, size, (FontStyle)style);
                        fi.Source = font_data;
                        font_data.Data = fi;
                        break;
                    }
                }
            }

            RegisterAction(new BootEngine(EndOfTheGame, null, ParseModule(rvii.GetData("Boot").AsString())));

            RenderLoop.Run(myForm, () => {
                long t1 = Stopwatch.GetTimestamp();
                Tick(new TimeSpan((t1 - myPrevTick) * TimeSpan.TicksPerSecond / Stopwatch.Frequency));
                Render(null);
                myPrevTick = t1;
                TimeSpan ts = new TimeSpan((Stopwatch.GetTimestamp() - t1) * TimeSpan.TicksPerSecond / Stopwatch.Frequency);
                System.Threading.Thread.Sleep(Math.Max(0, (int)(1000.0 / 60 - ts.TotalMilliseconds)));
            });
        }

        private void PrepareD3DResource()
        {
            mySprite = new Sprite(myDev);

            myPlaneTex = new Texture(myDev, 1, 1, 1, Usage.None, Format.A8R8G8B8, Pool.Managed);
            SharpDX.DataRectangle dr = myPlaneTex.LockRectangle(0, LockFlags.None);
            System.Runtime.InteropServices.Marshal.WriteInt32(dr.DataPointer, -1);
            myPlaneTex.UnlockRectangle(0);

            myShaderTex = new Texture(myDev, myD3DPP.BackBufferWidth, myD3DPP.BackBufferHeight, 1, Usage.RenderTarget, Format.X8R8G8B8, Pool.Default);

            myWndTex = new Texture(myDev, myD3DPP.BackBufferWidth, myD3DPP.BackBufferHeight, 1, Usage.RenderTarget, Format.A8R8G8B8, Pool.Default);
        }

        private void UnprepareD3DResource()
        {
            if (mySprite != null) {
                mySprite.Dispose();
                mySprite = null;
            }

            if (myPlaneTex != null) {
                myPlaneTex.Dispose();
                myPlaneTex = null;
            }

            if (myShaderTex != null) {
                myShaderTex.Dispose();
                myShaderTex = null;
            }

            if (myWndTex != null) {
                myWndTex.Dispose();
                myWndTex = null;
            }
        }

        public override void Extinguisher()
        {
            InputUtil.UnregisterKeyboard();

            Element[] elms = Element.GetElements("WaveBank");
            if (elms != null) {
                foreach (WaveInfo t in elms) {
                    if (t != null && t.Instance != null) {
                        t.Instance.Dispose();
                        t.Instance = null;
                    }
                }
            }

            if (mySoundBank != null) {
                mySoundBank.Dispose();
                mySoundBank = null;
            }

            if (myAudioEngine != null) {
                myAudioEngine.Dispose();
                myAudioEngine = null;
            }

            elms = Element.GetElements("Shader");
            if (elms != null) {
                foreach (ShaderInfo t in elms) {
                    if (t != null && t.Instance != null) {
                        t.Instance.Dispose();
                        t.Instance = null;
                    }
                }
            }

            Visual.ClearFontTexCache();

            elms = Element.GetElements("Texture");
            if (elms != null) {
                foreach (TextureInfo t in elms) {
                    if (t != null && t.Instance != null) {
                        t.Instance.Dispose();
                        t.Instance = null;
                    }
                }
            }

            UnprepareD3DResource();

            if (myDev != null) {
                myDev.Dispose();
                myDev = null;
            }

            if (myD3D != null) {
                myD3D.Dispose();
                myD3D = null;
            }

            if (myPFC != null) {
                myPFC.Dispose();
                myPFC = null;
            }

            if (myForm != null) {
                myForm.Close();
                myForm = null;
            }

            base.Extinguisher();
        }

        public override void Tick(TimeSpan delta)
        {
            if (!myDevMustReset) {
                InputUtil.PollState();
                myAudioEngine.DoWork();
                base.Tick(delta);
            } else {
                UnprepareD3DResource();
                if (myDev.TestCooperativeLevel() == SharpDX.Direct3D9.ResultCode.DeviceNotReset.Result.Code) {
                    try {
                        myDev.Reset(myD3DPP);
                        PrepareD3DResource();
                        myDevMustReset = false;
                    } catch {
                        System.Threading.Thread.Sleep(1000);
                    }
                }
            }
        }

        public override void Render(IGraphicsDevice r)
        {
            if (myDevMustReset) return;

            GraphicsObjects go = new GraphicsObjects(mySprite);

            myDev.BeginScene();
            try {
                Visual.TrimFontTexCache();
                Surface parentRT = myDev.GetRenderTarget(0);

                using (Surface shader_rt = myShaderTex.GetSurfaceLevel(0)) {
                    myDev.SetRenderTarget(0, shader_rt);
                    myDev.Clear(ClearFlags.Target, SharpDX.Color.Black, 1.0f, 0);
                    go.SpriteBatch.Begin(SpriteFlags.AlphaBlend | SpriteFlags.DoNotSaveState);
                    base.Render(go);
                    go.SpriteBatch.End();
                }

                myDev.SetRenderTarget(0, parentRT);
                go.SpriteBatch.Begin(SpriteFlags.DoNotSaveState);
                go.SpriteBatch.Draw(myShaderTex, SharpDX.Color.White);
                go.SpriteBatch.End();

                parentRT.Dispose();

            } finally {
                myDev.EndScene();
            }

            try {
                myDev.Present();
            } catch (SharpDX.SharpDXException sdxex) {
                if (sdxex.ResultCode == SharpDX.Direct3D9.ResultCode.DeviceLost.Result.Code) {
                    myDevMustReset = true;
                } else {
                    throw;
                }
            }
        }
    }
}
