﻿using System;
using System.Collections.Generic;
using System.Drawing;
using GustFront;
using SharpDX.Direct3D9;
using SharpDX.XACT3;
using App = RVII.RVIIManager;

namespace RVII
{
    public abstract class RVIIEngine : ScriptEngine
    {
        private ActionProcessInfo myFirstProcess = null;
        protected ActionProcessInfo FirstProcess
        {
            get { return myFirstProcess; }
        }

        protected RVIIEngine(UnregisterActionCallback unreg, ModuleInfo script, ScriptValue[] args)
            : base(unreg, App.Global)
        {
            myFirstProcess = BeginProcess(script, args);
        }

        #region "cmd impl"

        private ScriptValue cmdLoadData(ScriptValue[] @params)
        {
            return new ScriptValue(
                App.Global.Load(@params[0].AsInt32(), @params[1].AsString(), @params[2].AsString()));
        }

        private ScriptValue cmdSaveData(ScriptValue[] @params)
        {
            return new ScriptValue(
                App.Global.Save(@params[0].AsInt32(), @params[1].AsString(), @params[2].AsString()));
        }

        private ScriptValue cmdGetScreenSize(ScriptValue[] @params)
        {
            return new ScriptValue(new PointWrapper((Point)App.Global.GetViewBounds().Size));
        }

        private ScriptValue cmdToggleFullScreen(ScriptValue[] @params)
        {
            throw new NotSupportedException();
        }

        private ScriptValue cmdLoadTexture(ScriptValue[] @params)
        {
            TextureInfo ti = (TextureInfo)@params[0].Instance;
            if (ti.Instance == null) {
                ti.Instance = Texture.FromStream(App.Global.GraphicsDevice, App.Global.GetStreamToRead(ti.GetData("Path").AsString()), -1, -1, 1, Usage.None, Format.A8R8G8B8, Pool.Managed, Filter.None, Filter.None, 0);
                return true;
            } else {
                return false;
            }
        }

        private ScriptValue cmdUnloadTexture(ScriptValue[] @params)
        {
            TextureInfo ti = (TextureInfo)@params[0].Instance;
            if (ti.Instance != null) {
                ti.Instance.Dispose();
                ti.Instance = null;
                return true;
            } else {
                return false;
            }
        }

        private ScriptValue cmdGetShaderParam(ScriptValue[] @params)
        {
            Effect fx = ((ShaderInfo)@params[0].Instance).Instance;
            EffectHandle eh = fx.GetParameter(null, @params[1].AsString());
            ParameterDescription pd = fx.GetParameterDescription(eh);
            switch (pd.Type) {
                case ParameterType.Bool:
                    return fx.GetValue<bool>(eh);
                case ParameterType.Int:
                    return fx.GetValue<int>(eh);
                case ParameterType.Float:
                    return fx.GetValue<float>(eh);
                case ParameterType.String:
                    return fx.GetString(eh);
                default:
                    throw new Exception("取得できない種類のパラメータです。");
            }
        }

        private ScriptValue cmdSetShaderParam(ScriptValue[] @params)
        {
            Effect fx = ((ShaderInfo)@params[0].Instance).Instance;
            EffectHandle eh = fx.GetParameter(null, @params[1].AsString());
            ParameterDescription pd = fx.GetParameterDescription(eh);
            switch (pd.Type) {
                case ParameterType.Bool:
                    fx.SetValue<bool>(eh, @params[2].AsBoolean());
                    return true;
                case ParameterType.Int:
                    fx.SetValue<int>(eh, @params[2].AsInt32());
                    return true;
                case ParameterType.Float:
                    fx.SetValue<float>(eh, @params[2].AsSingle());
                    return true;
                case ParameterType.String:
                    fx.SetString(eh, @params[2].AsString());
                    return true;
                default:
                    throw new Exception("設定できない種類のパラメータです。");
            }
        }

        private ScriptValue cmdLoadWaveBank(ScriptValue[] @params)
        {
            WaveInfo wi = (WaveInfo)@params[0].Instance;
            if (wi.Instance == null) {
                if (wi.GetData("IsStreaming").AsBoolean() == false) {
                    byte[] bt = null;
                    using (System.IO.Stream strm = App.Global.GetStreamToRead(wi.GetData("Path").AsString())) {
                        bt = new byte[(int)strm.Length];
                        strm.Read(bt, 0, bt.Length);
                    }
                    wi.Instance = new WaveBank(App.Global.AudioEngine,
                        SharpDX.DataStream.Create<byte>(bt, true, false, 0, true));
                    //This is for avoiding current version SharpDX's bug.
                    //If you write following code and run it,
                    //WaveBank ctor reads all bytes from the file at Path.
                    //However the bytes will be discarded by next GC, and app will crash when you play it.
                    //Maybe this issue is fixed on next version.
                    //
                    //wi.Instance = new WaveBank(Engine.Global.AudioEngine,
                    //    System.IO.File.OpenRead(wi.GetData("Path").AsString()));
                } else {
                    wi.Instance = new WaveBank(App.Global.AudioEngine,
                        wi.GetData("Path").AsString(), wi.GetData("Offset").AsInt32(), wi.GetData("Packet").AsInt16());
                }
                for (int i = 0; i < 1000; i++) {
                    if ((wi.Instance.State & 0x00000004) == 0) {
                        App.Global.AudioEngine.DoWork();
                        System.Threading.Thread.Sleep(10);
                    } else {
                        return true;
                    }
                }
                throw new Exception("ウェーブバンクの読み込みがタイムアウトしました。");
            } else {
                return false;
            }
        }

        private ScriptValue cmdUnloadWaveBank(ScriptValue[] @params)
        {
            WaveInfo wi = (WaveInfo)@params[0].Instance;
            if (wi.Instance != null) {
                wi.Instance.Dispose();
                wi.Instance = null;
                return true;
            } else {
                return false;
            }
        }

        private ScriptValue cmdPlayAudio(ScriptValue[] @params)
        {
            short ci = App.Global.SoundBank.GetCueIndex(@params[0].AsString());
            Cue c = App.Global.SoundBank.Prepare(ci, 0);
            c.Play();
            return new ScriptValue(new CueWrapper(c));
        }

        private ScriptValue cmdSetVolume(ScriptValue[] @params)
        {
            App.Global.AudioEngine.SetVolume(App.Global.AudioEngine.GetCategory(@params[0].AsString()), @params[1].AsInt32() / 100.0f);
            return 0;
        }

        private ScriptValue cmdMapButtonToInput(ScriptValue[] @params)
        {
            InputUtil.MapButton(@params[0].AsInt32(), @params[1].AsInt32() - 1);
            return 0;
        }

        private ScriptValue cmdMapKeyToButton(ScriptValue[] @params)
        {
            InputUtil.MapKeyToButton(@params[0].AsInt32(), @params[1].AsInt32());
            return 0;
        }

        private ScriptValue cmdSetInputInterval(ScriptValue[] @params)
        {
            InputUtil.SetInputInterval(@params[0].AsInt32() - 1,
                @params[1].AsInt32(), @params[2].AsInt32());
            return 0;
        }

        private ScriptValue cmdGetInputState(ScriptValue[] @params)
        {
            return InputUtil.GetState();
        }

        private ScriptValue cmdGetPreviousInputState(ScriptValue[] @params)
        {
            return InputUtil.GetPreviousState();
        }

        protected override IDictionary<string, CommandImpl> GetCommandMap()
        {
            Dictionary<string, CommandImpl> dic = new Dictionary<string, CommandImpl>(base.GetCommandMap());

            dic.Add("LoadData", cmdLoadData);
            dic.Add("SaveData", cmdSaveData);

            dic.Add("GetScreenSize", cmdGetScreenSize);
            dic.Add("ToggleFullScreen", cmdToggleFullScreen);
            dic.Add("LoadTexture", cmdLoadTexture);
            dic.Add("UnloadTexture", cmdUnloadTexture);

            dic.Add("GetShaderParam", cmdGetShaderParam);
            dic.Add("SetShaderParam", cmdSetShaderParam);

            dic.Add("LoadWaveBank", cmdLoadWaveBank);
            dic.Add("UnloadWaveBank", cmdUnloadWaveBank);
            dic.Add("PlayAudio", cmdPlayAudio);
            dic.Add("SetVolume", cmdSetVolume);

            dic.Add("MapButtonToInput", cmdMapButtonToInput);
            dic.Add("MapKeyToButton", cmdMapKeyToButton);
            dic.Add("SetInputInterval", cmdSetInputInterval);
            dic.Add("GetInputState", cmdGetInputState);
            dic.Add("GetPreviousInputState", cmdGetPreviousInputState);

            return dic;
        }

        #endregion

        protected void ApplyFullScreenEffect(IGraphicsDevice r, Effect fx)
        {
            GraphicsObjects go = (GraphicsObjects)r;

            go.SpriteBatch.End();
            go.SpriteBatch.Begin(SpriteFlags.DoNotSaveState);
            int pass_count = fx.Begin(FX.DoNotSaveState);
            for (int i = 0; i < pass_count; i++) {
                fx.BeginPass(i);
                go.SpriteBatch.Draw(App.Global.ShaderTex, GSConv.Color(Color.White));
                fx.EndPass();
            }
            fx.End();
            go.SpriteBatch.End();
            go.SpriteBatch.Begin(SpriteFlags.AlphaBlend | SpriteFlags.DoNotSaveState);
        }
    }
}
